## $Id: customer.py 11604 2014-04-29 07:31:54Z henrik $
##
## Copyright (C) 2014 Uli Fouquet & Henrik Bettermann
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
##
"""
Container for the various objects owned by customers.
"""
import os
import re
import shutil
import grok
from datetime import datetime, timedelta
from hurry.workflow.interfaces import IWorkflowState, IWorkflowInfo
from zope.password.interfaces import IPasswordManager
from zope.component import getUtility, createObject
from zope.component.interfaces import IFactory
from zope.interface import implementedBy
from zope.securitypolicy.interfaces import IPrincipalRoleManager
from zope.schema.interfaces import ConstraintNotSatisfied

from waeup.ikoba.image import IkobaImageFile
from waeup.ikoba.imagestorage import DefaultFileStoreHandler
from waeup.ikoba.interfaces import (
    IObjectHistory, IUserAccount, IFileStoreNameChooser, IFileStoreHandler,
    IIkobaUtils, registration_states_vocab, IExtFileStore,
    CREATED, REQUESTED, APPROVED)
from waeup.ikoba.customers.export import EXPORTER_NAMES
from waeup.ikoba.customers.interfaces import (
    ICustomer, ICustomerNavigation, ICSVCustomerExporter)
from waeup.ikoba.customers.utils import generate_customer_id
from waeup.ikoba.customers.documents import CustomerDocumentsContainer
from waeup.ikoba.utils.helpers import attrs_to_fields, now, copy_filesystem_tree

RE_CUSTID_NON_NUM = re.compile('[^\d]+')


class Customer(grok.Container):
    """This is a customer container for the various objects
    owned by customers.
    """
    grok.implements(ICustomer, ICustomerNavigation)
    grok.provides(ICustomer)

    temp_password_minutes = 10

    def __init__(self):
        super(Customer, self).__init__()
        # The site doesn't exist in unit tests
        try:
            self.customer_id = generate_customer_id()
        except TypeError:
            self.customer_id = u'Z654321'
        self.password = None
        self.temp_password = None
        return

    def setTempPassword(self, user, password):
        """Set a temporary password (LDAP-compatible) SSHA encoded for
        officers.

        """
        passwordmanager = getUtility(IPasswordManager, 'SSHA')
        self.temp_password = {}
        self.temp_password[
            'password'] = passwordmanager.encodePassword(password)
        self.temp_password['user'] = user
        self.temp_password['timestamp'] = datetime.utcnow()  # offset-naive datetime

    def getTempPassword(self):
        """Check if a temporary password has been set and if it
        is not expired.

        Return the temporary password if valid,
        None otherwise. Unset the temporary password if expired.
        """
        temp_password_dict = getattr(self, 'temp_password', None)
        if temp_password_dict is not None:
            delta = timedelta(minutes=self.temp_password_minutes)
            now = datetime.utcnow()
            if now < temp_password_dict.get('timestamp') + delta:
                return temp_password_dict.get('password')
            else:
                # Unset temporary password if expired
                self.temp_password = None
        return None

    def writeLogMessage(self, view, message):
        ob_class = view.__implemented__.__name__.replace('waeup.ikoba.','')
        self.__parent__.logger.info(
            '%s - %s - %s' % (ob_class, self.__name__, message))
        return

    @property
    def display_fullname(self):
        middlename = getattr(self, 'middlename', None)
        ikoba_utils = getUtility(IIkobaUtils)
        return ikoba_utils.fullname(self.firstname, self.lastname, middlename)

    @property
    def fullname(self):
        middlename = getattr(self, 'middlename', None)
        if middlename:
            return '%s-%s-%s' % (self.firstname.lower(),
                middlename.lower(), self.lastname.lower())
        else:
            return '%s-%s' % (self.firstname.lower(), self.lastname.lower())

    @property
    def state(self):
        state = IWorkflowState(self).getState()
        return state

    @property
    def translated_state(self):
        state = registration_states_vocab.getTermByToken(
            self.state).title
        return state

    @property
    def history(self):
        history = IObjectHistory(self)
        return history

    @property
    def customer(self):
        return self

    @property
    def personal_data_expired(self):
        return False

# Set all attributes of Customer required in ICustomer as field
# properties. Doing this, we do not have to set initial attributes
# ourselves and as a bonus we get free validation when an attribute is
# set.
Customer = attrs_to_fields(Customer)


class CustomerFactory(grok.GlobalUtility):
    """A factory for customers.
    """
    grok.implements(IFactory)
    grok.name(u'waeup.Customer')
    title = u"Create a new customer.",
    description = u"This factory instantiates new customer instances."

    def __call__(self, *args, **kw):
        return Customer()

    def getInterfaces(self):
        return implementedBy(Customer)


@grok.subscribe(ICustomer, grok.IObjectAddedEvent)
def handle_customer_added(customer, event):
    """If a customer is added all subcontainers are automatically added
    and the transition create is fired. The latter produces a logging
    message.
    """
    # Assign global customer role for new customer
    account = IUserAccount(customer)
    account.roles = ['waeup.Customer']
    # Assign local CustomerRecordOwner role
    role_manager = IPrincipalRoleManager(customer)
    role_manager.assignRoleToPrincipal(
        'waeup.local.CustomerRecordOwner', customer.customer_id)
    if customer.state is None:
        IWorkflowInfo(customer).fireTransition('create')
    documents = CustomerDocumentsContainer()
    customer['documents'] = documents
    return


def path_from_custid(customer_id):
    """Convert a customer_id into a predictable relative folder path.

    Used for storing files.

    Returns the name of folder in which files for a particular customer
    should be stored. This is a relative path, relative to any general
    customers folder with 5 zero-padded digits (except when customer_id
    is overlong).

    We normally map 1,000 different customer ids into one single
    path. For instance ``K1000000`` will give ``01000/K1000000``,
    ``K1234567`` will give ``0123/K1234567`` and ``K12345678`` will
    result in ``1234/K12345678``.

    For lower numbers < 10**6 we return the same path for up to 10,000
    customer_ids. So for instance ``KM123456`` will result in
    ``00120/KM123456`` (there will be no path starting with
    ``00123``).

    Works also with overlong number: here the leading zeros will be
    missing but ``K123456789`` will give reliably
    ``12345/K123456789`` as expected.
    """
    # remove all non numeric characters and turn this into an int.
    num = int(RE_CUSTID_NON_NUM.sub('', customer_id))
    if num < 10**6:
        # store max. of 10000 custs per folder and correct num for 5 digits
        num = num / 10000 * 10
    else:
        # store max. of 1000 custs per folder
        num = num / 1000
    # format folder name to have 5 zero-padded digits
    folder_name = u'%05d' % num
    folder_name = os.path.join(folder_name, customer_id)
    return folder_name


def move_customer_files(customer, del_dir):
    """Move files belonging to `customer` to `del_dir`.

    `del_dir` is expected to be the path to the site-wide directory
    for storing backup data.

    The current files of the customer are removed after backup.

    If the customer has no associated files stored, nothing is done.
    """
    cust_id = customer.customer_id

    src = getUtility(IExtFileStore).root
    src = os.path.join(src, 'customers', path_from_custid(cust_id))

    dst = os.path.join(
        del_dir, 'media', 'customers', path_from_custid(cust_id))

    if not os.path.isdir(src):
        # Do not copy if no files were stored.
        return
    if not os.path.exists(dst):
        os.makedirs(dst, 0755)
    copy_filesystem_tree(src, dst)
    shutil.rmtree(src)
    return


def update_customer_deletion_csvs(customer, del_dir):
    """Update deletion CSV files with data from customer.

    `del_dir` is expected to be the path to the site-wide directory
    for storing backup data.

    Each exporter available for customers (and their many subobjects)
    is called in order to export CSV data of the given customer to csv
    files in the site-wide backup directory for object data (see
    DataCenter).

    Each exported row is appended a column giving the deletion date
    (column `del_date`) as a UTC timestamp.
    """
    for name in EXPORTER_NAMES:
        exporter = getUtility(ICSVCustomerExporter, name=name)
        csv_data = exporter.export_customer(customer)
        csv_data = csv_data.split('\r\n')

        # append a deletion timestamp on each data row
        timestamp = str(now().replace(microsecond=0))  # store UTC timestamp
        for num, row in enumerate(csv_data[1:-1]):
            csv_data[num+1] = csv_data[num+1] + ',' + timestamp
        csv_path = os.path.join(del_dir, '%s.csv' % name)

        # write data to CSV file
        if not os.path.exists(csv_path):
            # create new CSV file (including header line)
            csv_data[0] = csv_data[0] + ',del_date'
            open(csv_path, 'wb').write('\r\n'.join(csv_data))
        else:
            # append existing CSV file (omitting headerline)
            open(csv_path, 'a').write('\r\n'.join(csv_data[1:]))
    return


@grok.subscribe(ICustomer, grok.IObjectRemovedEvent)
def handle_customer_removed(customer, event):
    """If a customer is removed a message is logged and data is put
       into a backup location.

    The data of the removed customer is appended to CSV files in local
    datacenter and any existing external files (passport images, etc.)
    are copied over to this location as well.

    Documents in the file storage refering to the given customer are
    removed afterwards (if they exist). Please make no assumptions
    about how the deletion takes place. Files might be deleted
    individually (leaving the customers file directory intact) or the
    whole customer directory might be deleted completely.

    All CSV rows created/appended contain a timestamp with the
    datetime of removal in an additional `del_date` column.

    XXX: blocking of used customer_ids yet not implemented.
    """
    comment = 'Customer record removed'
    target = customer.customer_id
    try:
        site = grok.getSite()
        site['customers'].logger.info('%s - %s' % (
            target, comment))
    except KeyError:
        # If we delete an entire university instance there won't be
        # a customers subcontainer
        return

    del_dir = site['datacenter'].deleted_path

    # save files of the customer
    move_customer_files(customer, del_dir)

    # update CSV files
    update_customer_deletion_csvs(customer, del_dir)
    return

#: The file id marker for customer files
CUSTOMER_FILE_STORE_NAME = 'file-customer'


class CustomerFileNameChooser(grok.Adapter):
    """A file id chooser for :class:`Customer` objects.

    `context` is an :class:`Customer` instance.

    The :class:`CustomerImageNameChooser` can build/check file ids for
    :class:`Customer` objects suitable for use with
    :class:`ExtFileStore` instances. The delivered file_id contains
    the file id marker for :class:`Customer` object and the customer id
    of the context customer.

    This chooser is registered as an adapter providing
    :class:`waeup.ikoba.interfaces.IFileStoreNameChooser`.

    File store name choosers like this one are only convenience
    components to ease the task of creating file ids for customer
    objects. You are nevertheless encouraged to use them instead of
    manually setting up filenames for customers.

    .. seealso:: :mod:`waeup.ikoba.imagestorage`

    """
    grok.context(ICustomer)
    grok.implements(IFileStoreNameChooser)

    def checkName(self, name=None, attr=None):
        """Check whether the given name is a valid file id for the context.

        Returns ``True`` only if `name` equals the result of
        :meth:`chooseName`.

        """
        return name == self.chooseName()

    def chooseName(self, attr, name=None):
        """Get a valid file id for customer context.

        *Example:*

        For a customer with customer id ``'A123456'`` and
        with attr ``'nice_image.jpeg'`` stored in
        the customers container this chooser would create:

          ``'__file-customer__customers/A/A123456/nice_image_A123456.jpeg'``

        meaning that the nice image of this applicant would be
        stored in the site-wide file storage in path:

          ``customers/A/A123456/nice_image_A123456.jpeg``

        """
        basename, ext = os.path.splitext(attr)
        cust_id = self.context.customer_id
        marked_filename = '__%s__%s/%s_%s%s' % (
            CUSTOMER_FILE_STORE_NAME, path_from_custid(cust_id), basename,
            cust_id, ext)
        return marked_filename


class CustomerFileStoreHandler(DefaultFileStoreHandler, grok.GlobalUtility):
    """Customer specific file handling.

    This handler knows in which path in a filestore to store customer
    files and how to turn this kind of data into some (browsable)
    file object.

    It is called from the global file storage, when it wants to
    get/store a file with a file id starting with
    ``__file-customer__`` (the marker string for customer files).

    Like each other file store handler it does not handle the files
    really (this is done by the global file store) but only computes
    paths and things like this.
    """
    grok.implements(IFileStoreHandler)
    grok.name(CUSTOMER_FILE_STORE_NAME)

    def pathFromFileID(self, store, root, file_id):
        """All customer files are put in directory ``customers``.
        """
        marker, filename, basename, ext = store.extractMarker(file_id)
        sub_root = os.path.join(root, 'customers')
        return super(CustomerFileStoreHandler, self).pathFromFileID(
            store, sub_root, basename)

    def createFile(self, store, root, filename, file_id, file):
        """Create a browsable file-like object.
        """
        # call super method to ensure that any old files with
        # different filename extension are deleted.
        file, path, file_obj = super(
            CustomerFileStoreHandler, self).createFile(
            store, root,  filename, file_id, file)
        return file, path, IkobaImageFile(
            file_obj.filename, file_obj.data)
